import json
from typing import List
import pytest
from pymultirole_plugins.v1.schema import Document

from assertions import assert_result
from pyannotators_patterns.patterns import PatternsAnnotator, PatternsParameters


@pytest.fixture(scope="module")
def annotator():
    return PatternsAnnotator()


@pytest.fixture(scope="module")
def parameters():
    return PatternsParameters(mapping={
        "email": json.dumps({
            "patterns" : [
                {
                    "name": "Email (Medium)",
                    "regex": r"\b((([!#$%&'*+\-/=?^_`{|}~\w])|([!#$%&'*+\-/=?^_`{|}~\w][!#$%&'*+\-/=?^_`{|}~\.\w]{0,}[!#$%&'*+\-/=?^_`{|}~\w]))[@]\w+([-.]\w+)*\.\w+([-.]\w+)*)\b",
                    "score": 0.5,
                }
            ],
            "context": ["email"]
        })
    })


@pytest.mark.parametrize(
    "text, expected_len, expected_res",
    [
        # fmt: off
        # valid email addresses
        ("info@presidio.site", 1, ((0, 18),),),
        ("my email address is info@presidio.site", 1, ((20, 38),),),
        ("try one of these emails: info@presidio.site or anotherinfo@presidio.site",
            2,
         ((25, 43), (47, 72),),),
        # invalid email address
        ("my email is info@presidio.", 0, ()),
        # fmt: on
    ],
)
def test_when_all_email_addresses_then_succeed(
        text,
        expected_len,
        expected_res,
        annotator,
        parameters,
):
    entity_type = list(parameters.mapping.keys())[0]
    docs: List[Document] = annotator.annotate([Document(text=text, metadata={'language': 'en'})], parameters)
    doc0 = docs[0]
    assert len(doc0.annotations) == expected_len
    for res, (start, end) in zip(doc0.annotations, expected_res):
        assert_result(res, entity_type, start, end, None)



from typing import List

import pytest
from pymultirole_plugins.v1.schema import Document

from assertions import assert_result
from pyannotators_patterns.patterns import PatternsAnnotator, PatternsParameters, PATTERNS_EXAMPLE_STR


@pytest.fixture(scope="module")
def annotator():
    return PatternsAnnotator()


@pytest.fixture(scope="module")
def parameters():
    return PatternsParameters(mapping={
        "cc": PATTERNS_EXAMPLE_STR
    })


@pytest.mark.parametrize(
    "text, expected_len, expected_res",
    [
        # fmt: off
        (
            "4012888888881881 4012-8888-8888-1881 4012 8888 8888 1881",
            3, ((0, 16), (17, 36), (37, 56),),
        ),
        ("122000000000003", 1, ((0, 15),),),
        ("my credit card: 122000000000003", 1, ((16, 31),),),
        ("371449635398431", 1, ((0, 15),),),
        ("5555555555554444", 1, ((0, 16),),),
        ("5019717010103742", 1, ((0, 16),),),
        ("30569309025904", 1, ((0, 14),),),
        ("6011000400000000", 1, ((0, 16),),),
        ("3528000700000000", 1, ((0, 16),),),
        ("6759649826438453", 1, ((0, 16),),),
        ("5555555555554444", 1, ((0, 16),),),
        ("4111111111111111", 1, ((0, 16),),),
        ("4917300800000000", 1, ((0, 16),),),
        ("4484070000000000", 1, ((0, 16),),),
        # fmt: on
    ],
)
def test_when_all_credit_cards_then_succeed(
        text,
        expected_len,
        expected_res,
        annotator,
        parameters,
):
    entity_type = list(parameters.mapping.keys())[0]
    docs: List[Document] = annotator.annotate([Document(text=text, metadata={'language': 'en'})], parameters)
    doc0 = docs[0]
    assert len(doc0.annotations) == expected_len
    for res, (start, end) in zip(doc0.annotations, expected_res):
        assert_result(res, entity_type, start, end, None)



import json
from typing import List
import pytest
from pymultirole_plugins.v1.schema import Document

from assertions import assert_result
from pyannotators_patterns.patterns import PatternsAnnotator, PatternsParameters


@pytest.fixture(scope="module")
def annotator():
    return PatternsAnnotator()


@pytest.fixture(scope="module")
def parameters():
    return PatternsParameters(mapping={
        "zip": json.dumps({
            "patterns" : [
                {
                    "name": "zip code (weak)",
                    "regex": "(\\b\\d{5}(?:\\-\\d{4})?\\b)",
                    "score": 0.01,
                }
            ],
            "context": ["zip", "code"]
        })
    })


@pytest.mark.parametrize(
    "text, expected_len, expected_res",
    [
        # fmt: off
        # valid email addresses
        ("My zip code is 90210", 1, ((15, 20),),),
        # fmt: on
    ],
)
def test_when_all_email_addresses_then_succeed(
        text,
        expected_len,
        expected_res,
        annotator,
        parameters,
):
    entity_type = list(parameters.mapping.keys())[0]
    docs: List[Document] = annotator.annotate([Document(text=text, metadata={'language': 'en'})], parameters)
    doc0 = docs[0]
    assert len(doc0.annotations) == expected_len
    for res, (start, end) in zip(doc0.annotations, expected_res):
        assert_result(res, entity_type, start, end, None)



import json
import logging
from functools import lru_cache
from typing import Type, List, cast, Dict

import spacy
from presidio_analyzer import Pattern, PatternRecognizer, RecognizerRegistry, AnalyzerEngine, LemmaContextAwareEnhancer
from presidio_analyzer.nlp_engine import SpacyNlpEngine
from pydantic import BaseModel, Field
from pymultirole_plugins.util import comma_separated_to_list
from pymultirole_plugins.v1.annotator import AnnotatorParameters, AnnotatorBase
from pymultirole_plugins.v1.schema import Document, Span, Annotation
from spacy.cli.download import download_model, get_compatibility, get_version
from spacy.language import Language
from wasabi import msg

logger = logging.getLogger(__name__)

PATTERNS_EXAMPLE = {
    "patterns": [
        {
            "name": "All Credit Cards (weak)",
            "regex": r"\b((4\d{3})|(5[0-5]\d{2})|(6\d{3})|(1\d{3})|(3\d{3}))[- ]?(\d{3,4})[- ]?(\d{3,4})[- ]?(\d{3,5})\b",
            "score": 0.3
        }
    ],
    "context": [
        "credit",
        "card",
        "visa",
        "mastercard",
        "cc ",
        "amex",
        "discover",
        "jcb",
        "diners",
        "maestro",
        "instapayment",
    ]
}
PATTERNS_EXAMPLE_STR = json.dumps(PATTERNS_EXAMPLE, indent=2)


class PatternsParameters(AnnotatorParameters):
    mapping: Dict[str, str] = Field(None,
                                    description="""List of regex patterns to be used by the annotator and optionnal list of context words to increase confidence in detection.<br/>
                            Each pattern is composed of a name, a list of python regular expressions and a score indicating the pattern's strength (values varies from 0 to 1).<br/>
                            For example you can define a regex pattern to recognize credit card numbers addresses along with some [context words](https://microsoft.github.io/presidio/tutorial/06_context/) like this:<br/>
    ```""" + PATTERNS_EXAMPLE_STR + "```", extra="key:label,val:json")

    score_threshold: float = Field(0.0, description="Minimum confidence value for detected entities to be returned")
    context_similarity_factor: float = Field(0.0,
                                             description="How much to enhance confidence of match entity, as explained [here](https://microsoft.github.io/presidio/tutorial/06_context/)",
                                             extra="advanced")
    min_score_with_context_similarity: float = Field(0.4,
                                                     description="Minimum confidence score, as explained [here](https://microsoft.github.io/presidio/tutorial/06_context/)",
                                                     extra="advanced")
    context_prefix_count: int = Field(5,
                                      description="How many words before the entity to match context, as explained [here](https://microsoft.github.io/presidio/tutorial/06_context/)",
                                      extra="advanced")
    context_suffix_count: int = Field(0,
                                      description="How many words after the entity to match context, as explained [here](https://microsoft.github.io/presidio/tutorial/06_context/)",
                                      extra="advanced")


SUPPORTED_LANGUAGES = "en,fr,de,nl,es,pt,it,zh,ar,ru"


class PatternsAnnotator(AnnotatorBase):
    """Patterns annotator using the Presidio [regex pattern recognizer](https://microsoft.github.io/presidio/tutorial/02_regex/).
    Regular expressions can be tested [here](https://regex101.com/).
    #need-segments
    #languages:""" + SUPPORTED_LANGUAGES

    def annotate(
            self, documents: List[Document], parameters: AnnotatorParameters
    ) -> List[Document]:
        params: PatternsParameters = cast(PatternsParameters, parameters)
        supported_languages = comma_separated_to_list(SUPPORTED_LANGUAGES)

        mapping = frozenset(params.mapping.items())
        labels = list(params.mapping.keys())
        registry = get_registry(mapping)

        for document in documents:
            # Retrieve nlp pipe
            lang = document_language(document, None)
            if lang is None or lang not in supported_languages:
                raise AttributeError(f"Metadata language {lang} is required and must be in {SUPPORTED_LANGUAGES}")
            nlp = get_nlp(lang)
            analyzer = AnalyzerEngine(registry=registry,
                                      nlp_engine=LoadedSpacyNlpEngine(lang, nlp),
                                      default_score_threshold=0,
                                      supported_languages=supported_languages,
                                      context_aware_enhancer=LemmaContextAwareEnhancer(params.context_similarity_factor,
                                                                                       params.min_score_with_context_similarity,
                                                                                       params.context_prefix_count,
                                                                                       params.context_suffix_count))

            document.annotations = []
            if not document.sentences:
                document.sentences = [Span(start=0, end=len(document.text))]

            for s in document.sentences:
                if s.end > s.start:
                    stext = document.text[s.start: s.end]
                    results = analyzer.analyze(text=stext, entities=labels,
                                               language=lang, return_decision_process=True,
                                               score_threshold=params.score_threshold)  # noqa D501
                    for result in results:
                        start = s.start + result.start
                        end = s.start + + result.end
                        document.annotations.append(
                            Annotation(
                                start=start,
                                end=end,
                                text=document.text[start: end],
                                labelName=result.entity_type,
                                score=result.score,
                                properties={"analysis_explanation": result.analysis_explanation}
                            )
                        )

            return documents

    @classmethod
    def get_model(cls) -> Type[BaseModel]:
        return PatternsParameters


def document_language(doc: Document, default: str = None):
    if doc.metadata is not None and 'language' in doc.metadata:
        return doc.metadata['language']
    return default


# Deprecated model shortcuts, only used in errors and warnings
MODEL_SHORTCUTS = {
    "en": "en_core_web_sm", "de": "de_core_news_sm", "es": "es_core_news_sm",
    "pt": "pt_core_news_sm", "fr": "fr_core_news_sm", "it": "it_core_news_sm",
    "nl": "nl_core_news_sm", "el": "el_core_news_sm", "nb": "nb_core_news_sm",
    "lt": "lt_core_news_sm", "xx": "xx_ent_wiki_sm", "zh": "zh_core_web_sm",
    "ru": "ru_core_news_sm"
}


@lru_cache(maxsize=None)
def get_nlp(lang):
    model = MODEL_SHORTCUTS.get(lang, lang)
    try:
        nlp: Language = spacy.load(model)
    except BaseException:
        nlp = load_spacy_model(model)
    return nlp


# Create a class inheriting from SpacyNlpEngine
class LoadedSpacyNlpEngine(SpacyNlpEngine):
    def __init__(self, lang, loaded_spacy_model):
        self.nlp = {lang: loaded_spacy_model}


@lru_cache(maxsize=None)
def get_registry(mapping_items):
    recognizers = []
    for pname, pvalue in mapping_items:
        pattern_definition = json.loads(pvalue)
        patterns = [Pattern.from_dict(pat) for pat in pattern_definition['patterns']]
        recognizer = PatternRecognizer(
            supported_entity=pname, patterns=patterns, context=pattern_definition.get('context', None)
        )
        recognizers.append(recognizer)
    registry = RecognizerRegistry(recognizers)
    return registry


def load_spacy_model(model, *pip_args):
    suffix = "-py3-none-any.whl"
    dl_tpl = "{m}-{v}/{m}-{v}{s}#egg={m}=={v}"
    model_name = model
    if model in MODEL_SHORTCUTS:
        msg.warn(
            f"As of spaCy v3.0, shortcuts like '{model}' are deprecated. Please "
            f"use the full pipeline package name '{MODEL_SHORTCUTS[model]}' instead."
        )
        model_name = MODEL_SHORTCUTS[model]
    compatibility = get_compatibility()
    version = get_version(model_name, compatibility)
    download_model(dl_tpl.format(m=model_name, v=version, s=suffix), pip_args)
    msg.good(
        "Download and installation successful",
        f"You can now load the package via spacy.load('{model_name}')",
    )
    # If a model is downloaded and then loaded within the same process, our
    # is_package check currently fails, because pkg_resources.working_set
    # is not refreshed automatically (see #3923). We're trying to work
    # around this here be requiring the package explicitly.
    require_package(model_name)
    return spacy.load(model_name, exclude=["parser"])


def require_package(name):
    try:
        import pkg_resources

        pkg_resources.working_set.require(name)
        return True
    except:  # noqa: E722
        return False



error = 0.00001


def __assert_result_without_score(
    result, expected_entity_type, expected_start, expected_end
):
    assert result.labelName == expected_entity_type
    assert result.start == expected_start
    assert result.end == expected_end


def assert_result(
    result, expected_entity_type, expected_start, expected_end, expected_score
):
    __assert_result_without_score(
        result, expected_entity_type, expected_start, expected_end
    )
    if expected_score is not None:
        assert result.score == expected_score


def assert_result_within_score_range(
    result,
    expected_entity_type,
    expected_start,
    expected_end,
    expected_score_min,
    expected_score_max,
):
    __assert_result_without_score(
        result, expected_entity_type, expected_start, expected_end
    )
    min_score = max(0, expected_score_min - error)
    max_score = min(1, expected_score_max + error)
    assert result.score >= min_score and result.score <= max_score
